"""
A module for file-related functions.
"""

from __future__ import annotations

import codecs
import csv
import hashlib
import io
import os
import tempfile
from typing import IO, Any, AnyStr, Callable


XLRD_FLOAT_TYPE = 2


def size(file: IO[AnyStr]) -> int:
    """
    Return the size of the file in bytes.
    """
    position = file.tell()
    file.seek(0, os.SEEK_END)
    size_ = file.tell()
    file.seek(position)
    return size_


def hashfile(
    file_handle: IO[bytes],
    hasher: Callable[..., hashlib._Hash] = hashlib.sha256,
    blocksize: int = 65536,
) -> str:
    """
    Get contents of file as hash. Use like so:
    >>> hashfile(open(myfile, 'rb'))
    ... 40c33eec0a307e7726992d3786f11a8c30309489455c33cfc14b36e6fbf7347b
    """
    hash_ = hasher()
    _buffer = file_handle.read(blocksize)
    while len(_buffer) > 0:
        hash_.update(_buffer)
        _buffer = file_handle.read(blocksize)
    return hash_.hexdigest()


def convert_xlsx_to_csv(
    xlsx_filepath: str,
    csv_filepath: str | None = None,
    encoding: str | None = None,
    errors: str | None = None,
    quoting: int | None = csv.QUOTE_ALL,
    delimiter: str | None = ",",
) -> IO[str]:
    """
    Convert .xlsx files to CSV format using `openpyxl` and return a handle to the output file.

    Args:
        xlsx_filepath:     Path of .xls or .xlsx file to convert.
        csv_filepath:       Location of result file referenced by returned handler. A temporary
                            file will be created if this arg is not present.
        encoding:           Encoding scheme for csv file.
        errors:             How file object should handle encoding errors.
        quoting:            The type of quoting the output CSV file should have
        delimiter:          The delimiter the output CSV file should have

    """
    # data_only to extract values from formulae and not formulas themselves
    import openpyxl

    workbook = openpyxl.load_workbook(xlsx_filepath, data_only=True, read_only=True)
    sheet = workbook.active

    csv_file, wr = _get_csv_file_and_writer(
        csv_filepath, encoding, errors, quoting, delimiter
    )

    for row in sheet.rows:
        wr.writerow([cell.value for cell in row])

    # Return handle to csv object
    csv_file.seek(0)
    return csv_file


def convert_xls_to_csv(
    xls_filepath: str,
    csv_filepath: str | None = None,
    encoding: str | None = None,
    errors: str | None = None,
    quoting: int | None = csv.QUOTE_ALL,
    delimiter: str | None = ",",
) -> IO[str]:
    """
    Convert .xls files to CSV format using `xlrd` and return a handle to the output file.

    Args:
        xls_filepath:       Path of .xls file to convert.
        csv_filepath:       Location of result file referenced by returned handler. A temporary
                            file will be created if this arg is not present.
        encoding:           Encoding scheme for csv file.
        errors:             How file object should handle encoding errors.
        quoting:            The type of quoting the CSV file should have
        delimiter:          The delimiter the output CSV file should have

    """
    import xlrd

    workbook = xlrd.open_workbook(xls_filepath)
    sheet = workbook.sheet_by_index(0)

    csv_file, wr = _get_csv_file_and_writer(
        csv_filepath, encoding, errors, quoting, delimiter
    )
    for rownum in range(sheet.nrows):
        row = sheet.row(rownum)
        values = []
        for cell in row:
            value = cell.value
            # Excel treats all numbers as floats
            if cell.ctype == XLRD_FLOAT_TYPE:
                if cell.value % 1 == 0.0:
                    value = int(value)
            values.append(value)
        wr.writerow(values)

    # Return handle to csv object
    csv_file.seek(0)
    return csv_file


def _get_csv_file_and_writer(
    csv_filepath: str | None,
    encoding: str | None,
    errors: str | None,
    quoting: int | None = csv.QUOTE_ALL,
    delimiter: str | None = ",",
) -> tuple[IO[str], Any]:
    if quoting is None:
        quoting = csv.QUOTE_ALL
    if delimiter is None:
        delimiter = ","

    if csv_filepath:
        csv_file: IO[str] = open(
            csv_filepath, mode="w+", encoding=encoding, errors=errors
        )
    else:
        # `error' argument added in 3.8
        csv_file = tempfile.NamedTemporaryFile(mode="w+", encoding=encoding)

    return csv_file, csv.writer(csv_file, quoting=quoting, delimiter=delimiter)


def remove_bom(string: str) -> str:
    """
    Remove the BOM mark from the beginning of a string if it exists.

    This mark is sometimes added to excel files and causes issues when reading.
    """
    return string[3:] if string.startswith(str(codecs.BOM_UTF8)) else string


def generate_csv_bytes(data: list[list[Any]], headers: list[Any]) -> bytes:
    """
    Generates a csv file and transforms it to bytes.
    reference: https://okhlopkov.medium.com/dont-save-a-file-on-disk-to-send-it-with-telegram-bot-d7cd591fec2d
    :param data: List containing the data rows
    :param headers: List containing the headers of the file
    """
    rows = [headers, *data]

    string_io = io.StringIO()

    csv.writer(string_io).writerows(rows)
    string_io.seek(0)

    csv_bytes = io.BytesIO()
    csv_bytes.write(string_io.getvalue().encode())
    csv_bytes.seek(0)

    return csv_bytes.read()


def convert_csv_to_parquet(
    data: bytes,
    na_values: str | dict[str, Any] | None = None,
    data_type: str | dict[str, Any] | None = None,
    parse_dates: list[str] | None = None,
) -> bytes:
    """
    Converts bytes from CSV format to Parquet.

    :param data: bytes contents for/from a CSV file
    :param na_values: Additional strings to recognize as NA/NaN. If dict passed, specific
    per-column NA values.
    :param data_type: Data type for data or columns. You can Use str, object or a dict with
    the column names and types. E.g. {‘a’: np.float64, ‘b’: np.int32, ‘c’: ‘Int64’}
    :return: bytes for a Parquet file
    """
    import pandas as pd

    dataframe = pd.read_csv(
        io.BytesIO(data),
        dtype=data_type,
        na_values=na_values,
        parse_dates=parse_dates,
        thousands=",",
    )
    bytes_io = io.BytesIO()
    dataframe.to_parquet(bytes_io, engine="pyarrow")
    bytes_io.seek(0)
    return bytes_io.read()



import pytest

from xocto import urls


class TestPopURLQueryParam:
    def test_returns_modifed_url_and_popped_value(self):
        test_url = "https://localhost/x/y/x?a=1&b=2&c=3"
        (new_url, param) = urls.pop_url_query_param(test_url, "b")
        assert new_url == "https://localhost/x/y/x?a=1&c=3"
        assert param == "2"

    def test_returns_unmodified_url_for_missing_values(self):
        """
        Missing values return an unchanged URL and None as the attribute value.
        """
        test_url = "https://localhost/x/y/x?a=1&c=3"
        (new_url, param) = urls.pop_url_query_param(test_url, "b")
        assert new_url == test_url
        assert param is None

    def test_keeps_slashes_on_schema_only_urls(self):
        """
        URLs which have no path or netloc don't lose their slashes.

        Without special handling in `pop_url_query_param`, schema only URLs would lose their
        trailing slashes.

        >>> urllib.parse.urlunparse(urllib.parse.urlparse("schema://"))
        "schema:"

        Maybe that's a "valid" and "equivalent" URL but some libraries check for the "://" string
        to identify a URL. So keeping the trailing slashes is required for compatability. One
        specific library which has this behaviour is pyFilesystem where `open_fs("mem://")` opens
        an in-memory filesystem and `open_fs("mem:")` opens the `./mem:` directory.
        """
        test_url = "mem://?a=1"
        (new_url, param) = urls.pop_url_query_param(test_url, "a")
        assert new_url == "mem://"
        assert param == "1"

        test_url = "mem://"
        (new_url, param) = urls.pop_url_query_param(test_url, "a")
        assert new_url == test_url
        assert param is None

    def test_keeps_slashes_on_schema_only_urls_with_multiple_variables(self):
        """
        URLs which have no path or netloc don't lose their slashes.

        Without special handling in `pop_url_query_param`, schema only URLs would lose their
        trailing slashes.

        >>> urllib.parse.urlunparse(urllib.parse.urlparse("schema://"))
        "schema:"

        Maybe that's a "valid" and "equivalent" URL but some libraries check for the "://" string
        to identify a URL. So keeping the trailing slashes is required for compatability. One
        specific library which has this behaviour is pyFilesystem where `open_fs("mem://")` opens
        an in-memory filesystem and `open_fs("mem:")` opens the `./mem:` directory.
        """
        test_url = "mem://?a=1&b=2"
        (new_url, param) = urls.pop_url_query_param(test_url, "a")
        assert new_url == "mem://?b=2"
        assert param == "1"

        test_url = "mem://?b=2"
        (new_url, param) = urls.pop_url_query_param(test_url, "a")
        assert new_url == test_url
        assert param is None

    def test_raises_value_error_for_params_with_multiple_values(self):
        """
        Raises a ValueError if the requested param has multiple values.
        """
        with pytest.raises(ValueError):
            urls.pop_url_query_param("https://localhost/x/y/x?a=1&a=2", "a")


class TestParseFileDestinationFromURL:
    def test_handles_upload_dir_being_subdir_of_destination_dir(self):
        """
        Where `upload` starts with no dots or slashes it is a subdirs of the destination dir.

        In that case the returned URL is that of the destination and the upload path is given
        relative to that.
        """
        dest_url = "ftp://some_server/gas/destination?upload=nested/sub/dir"
        (fs_url, dest_path, upload_path) = urls.parse_file_destination_from_url(
            dest_url
        )

        assert fs_url == "ftp://some_server/gas/destination"
        assert dest_path == "."
        assert upload_path == "nested/sub/dir"

    def test_handles_destination_dir_being_subdir_of_upload_dir(self):
        """
        Where `upload` is ".." it is the parent of the destination dir.

        In that case the returned URL is that of the upload dir and the destination path is given
        relative to that.
        """
        dest_url = "ftp://some_server/gas/destination?upload=.."
        (fs_url, dest_path, upload_path) = urls.parse_file_destination_from_url(
            dest_url
        )

        assert fs_url == "ftp://some_server/gas"
        assert dest_path == "destination"
        assert upload_path == "."

    def test_handles_destination_dir_and_upload_dir_being_cousins(self):
        """
        Where `upload` starts "..", destination and upload dirs will have common ancestor.

        The destination and upload dir can be in different parts of the file system with neither
        being a sub dir of the other. In that case the returned URL is that of the nearest common
        ancestor between the two dirs and both paths are given relative to that.
        """
        dest_url = "ftp://some_server/gas/destination/dir?upload=../../upload/sub/dir"
        (fs_url, dest_path, upload_path) = urls.parse_file_destination_from_url(
            dest_url
        )

        assert fs_url == "ftp://some_server/gas"
        assert dest_path == "destination/dir"
        assert upload_path == "upload/sub/dir"

    def test_handles_absolute_upload_dir(self):
        """
        Where `upload` starts "/", it is an absolute path.

        In that case the upload dir is resolved relative to the root of the FS URL.
        """
        dest_url = "ftp://some_server/gas/destination?upload=/gas/upload"
        (fs_url, dest_path, upload_path) = urls.parse_file_destination_from_url(
            dest_url
        )

        assert fs_url == "ftp://some_server/gas"
        assert dest_path == "destination"
        assert upload_path == "upload"

    def test_setting_destination(self):
        url = "ftp://oeftp:blahblah@172.30.195.140?destination=MAM/TO_OE"
        result = urls.parse_file_destination_from_url(url)

        assert result == (
            "ftp://oeftp:blahblah@172.30.195.140",
            "MAM/TO_OE",
            ".",
        )

    def test_setting_destination_and_upload(self):
        url = "ftp://oeftp:blahblah@172.30.195.140?destination=MAM/TO_OE&upload=MAM/TO_OE/pending"
        result = urls.parse_file_destination_from_url(url)

        assert result == (
            "ftp://oeftp:blahblah@172.30.195.140",
            "MAM/TO_OE",
            "MAM/TO_OE/pending",
        )



from typing import Any, Protocol, TypeVar


T = TypeVar("T", covariant=True)


class Comparable(Protocol[T]):
    """
    A way of describing an object that can be compared to another using some basic operations.
    """

    def __eq__(self, other: Any) -> bool: ...

    def __lt__(self, other: Any) -> bool: ...

    def __le__(self, other: Any) -> bool: ...

    def __gt__(self, other: Any) -> bool: ...

    def __ge__(self, other: Any) -> bool: ...



from __future__ import annotations

import os
from urllib import parse


def pop_url_query_param(url: str, key: str) -> tuple[str, str | None]:
    """
    Pop one query string value off a URL returning both the modified URL and popped value.

    E.g.
    >>> pop_url_query_param('https://localhost/x/y/x?a=1&b=2&c=3', 'b')
    ('https://localhost/x/y/x?a=1&c=3', '2')

    Missing values return an unchanged URL and None as the attribute value.
    >>> pop_url_query_param('https://localhost/x/y/x?a=1&b=2&c=3', 'd')
    ('https://localhost/x/y/x?a=1&b=2&c=3', None)

    Raises a ValueError if the requested param has multiple values.
    >>> pop_url_query_param('https://localhost/x/y/x?a=1&a=2', 'a')
    Traceback (most recent call last):
        ...
    ValueError: ...
    """
    parsed_url = parse.urlparse(url)
    query_dict = parse.parse_qs(
        parsed_url.query, keep_blank_values=True, errors="strict"
    )
    query_value_list = query_dict.pop(key, (None,))
    if len(query_value_list) != 1:
        raise ValueError(f"Cannot pop multi-valued query param: Key {key!r} in {url!r}")
    modified_url = parsed_url._replace(query=parse.urlencode(query_dict, doseq=True))
    resulting_url = parse.urlunparse(modified_url)
    resulting_url = _fix_url_scheme(old_url=url, new_url=resulting_url)
    return resulting_url, query_value_list[0]


def parse_file_destination_from_url(url: str) -> tuple[str, str, str]:
    """
    Split a destination URL into a pyFileSystem URL, destination path and upload path.

    A common problem when uploading files to file-systems like FTP or SFTP is that uploads are not
    atomic. That means that when uploading a file for another party to download then is a chance
    that they will see a partially uploaded file. To avoid that problem it's common to upload the
    file to an intermediary (upload) path and then move it to the final (destination) path. That
    approach works because while the upload is not atomic, the server-side rename/move is.

    When configuring pyFileSystems it's nice to use a single URL rather than separate values for
    HOST, PORT, PATH, USERNAME, PASSWORD, etc. This function extends those ergonomics by allowing a
    single URL to specific both the upload and final destination locations.

    The "Destination URLs" taken by this function are pyFileSystem URLs describing a directory into
    which a file should be written. This URL can have an optional `upload` key in the query string.
    If that key exists then is is a path to a directory relative to the main URL. This indicates
    where the file should be uploaded before being moved to the final destination

    This function parses the given "destination URL" into three parts.
    - A pyFileSystem URL which contains both the destination and upload directories
    - The destination path (Relative to the returned FS URL)
    - The upload path (Relative to the returned FS URL)

    pyFileSystem filesystems are sandboxed, which means we can't escape them using `..` or similar.
    That means that we have to ensure that the returned URL contains both the destination and
    upload paths. That means modifying the URL path to be the common ancestor of the two dirs.

    The `upload` dir can be a subdir of the destination` ...
    >>> parse_file_destination_from_url('ftp://some_server/dir/destination?upload=temp')
    'ftp://some_server/dir/destination', '.', 'temp')

    ... or a parent dir ...
    >>> parse_file_destination_from_url('ftp://some_server/dir/destination?upload=..')
    ('ftp://some_server/dir', 'destination', '.')

    ... Or a sibling ..
    >>> parse_file_destination_from_url('ftp://some_server/dir/destination?upload=../temp')
    ('ftp://some_server/dir', 'destination', 'temp')

    ... Or an absolute path.
    >>> parse_file_destination_from_url('ftp://some_server/dir/destination?upload=/temp')
    ('ftp://some_server/', 'dir/destination', 'temp')

    If the upload path is not given then it's assumed that the file should be directly uploaded to
    the destination dir.
    >>> parse_file_destination_from_url('ftp://some_server/dir/destination')
    ('ftp://some_server/dir/destination', '.', '.')

    A destination path can be specified with a `destination` query key. This is useful if you need
    to force a certain pyFileSystem URL rather than using the full path to the final `destination`.
    (c.f. the above example)
    >>> parse_file_destination_from_url('ftp://some_server/?destination=dir/destination')
    ('ftp://some_server/', 'dir/destination', '.')

    If both `destination` and `upload` query key are specified then they are both used. This is
    useful if you need to force a certain pyFileSystem URL rather than finding the nearest common
    ancestor of `destination` and `upload`
    >>> parse_file_destination_from_url('ftp://some_server/?destination=dir/destination&upload=dir/destination/temp')
    ('ftp://some_server/', 'dir/destination', 'dir/destination/temp')
    """
    (url, url_relative_upload_path) = pop_url_query_param(url, key="upload")
    (url, url_relative_destination_path) = pop_url_query_param(url, key="destination")

    if not url_relative_upload_path and url_relative_destination_path:
        return url, url_relative_destination_path, "."
    elif url_relative_upload_path and url_relative_destination_path:
        # Explicitly set destination and upload paths means we should use them without finding defaults.
        return url, url_relative_destination_path, url_relative_upload_path
    elif not url_relative_upload_path:
        return url, ".", "."
    else:
        parsed_url = parse.urlparse(url)
        destination_path = parsed_url.path
        upload_path = os.path.abspath(
            os.path.join(destination_path, url_relative_upload_path)
        )
        common_path = os.path.commonpath((destination_path, upload_path))
        new_url = parse.urlunparse(parsed_url._replace(path=common_path))
        new_url = _fix_url_scheme(old_url=url, new_url=new_url)
        upload_path = os.path.abspath(
            os.path.join(destination_path, url_relative_upload_path)
        )
        rel_destination_path = os.path.relpath(destination_path, common_path)
        rel_upload_path = os.path.relpath(upload_path, common_path)
        return new_url, rel_destination_path, rel_upload_path


# Private


def _fix_url_scheme(*, old_url: str, new_url: str) -> str:
    """
    Restore '://' to a URL where urlunparse(urlparse(url)) has removed it.

    Strictly speaking URLs which don't have a username, password or host don't need the "//" after
    the `schema:`. Python's urlunparse normalises URLs and "may result in a slightly different, but
    equivalent URL". This leads to results like:

    - urlunparse(urlparse("schema://")) == "schema:".
    - urlunparse(urlparse("schema:///root/subdir")) == "schema:/root/subdir"

    Unfortunately not all tools cope well with URLs that don't include the '://'. This function
    restores the "://" where a trip through urlunparse has removed it.
    """
    if "://" in old_url and "://" not in new_url:
        segments = new_url.split(":", maxsplit=1)
        new_url = segments[0] + "://" + segments[1]
    return new_url
